# -*- coding: utf-8 -*-
# Authors : Alix Chagué <alix.chague@inria.fr>
# Licence : MIT
"""
    The ``Parser Page`` module parse XML PAGE
    =========================================

    inspiration from:
    https://gitlab.inria.fr/scripta/escriptorium/-/blob/master/app/apps/imports/parsers.py
    ...

"""

import os

from bs4 import BeautifulSoup

from kami.kamutils._utils import (_report_log)


def _clean_coords(baseline_tag) -> list:
    """Clean coordinates following Kraken's expectation

    Example:

    >>> _clean_coords(BeautifulSoup('<Baseline points="645,2398 1318,2387"/>'))
    [[645.0, 2398.0], [1318.0, 2387.0]]

    :param baseline_tag: <Coords> node
    :type baseline_tag: BeautifulSoup
    :return: list of points
    :rtype: list
    """
    return [list(map(lambda x: 0 if float(x) < 0 else float(x), pt.split(",")))
            for pt in baseline_tag.get("points").split(" ")]


class _PageParser:
    """A PAGE XML Parser for KaMI.

    Attributes:
        file_name       Source file name if there is one
        page_content    XML Tree parsed with BeautifulSoup
        _pcgts          True if XML Tree has a PcGts node
        bounds          List of bounds for Kraken, 1 bound per segment
        transcriptions  List of transcriptions, 1 string per segment
        pairs           List of pairs of bounds and transcriptions, 1 dict per pair
        source_image    Source image referred to in //Page/@imageFilename

    ---

    PageParser.bounds[1] matches PageParser.transcriptions[1] and PageParser.pairs[1]
    """

    def _get_page_content(self, page):
        """Process PageParser input variable to obtain parsed PAGE XML tree

        Example:
        >>> _get_page_content("incorrect/path/to/unexisting/file.txt")
        None

        >>> type(_get_page_content("path/to/existing/file.xml"))
        <class 'bs4.BeautifulSoup'>
        >>> type(_get_page_content(BeautifulSoup('<example/>', 'xml')))
        <class 'bs4.BeautifulSoup'>

        :param page: either a path (str) either a parsed XML Tree
        :type page: str | BeautifulSoup
        :return: Parsed XML tree
        :rtype: BeautifulSoup
        """
        if isinstance(page, type(BeautifulSoup(features='xml'))):
            self.page_content = page
        elif isinstance(page, str):
            if os.path.isfile(page) and not page.endswith('.xml'):
                _report_log("PageParser expects PAGE XML content or a .xml file.", "E")
            elif os.path.isfile(page):
                try:
                    with open(page, 'r', encoding='utf8') as fh:
                        file_content = fh.read()
                except OSError as e:
                    _report_log("Something went wrong while parsing source file:", "E")
                    print(e)
                else:
                    try:
                        self.page_content = BeautifulSoup(file_content, 'xml')
                    except TypeError as e:
                        _report_log("Something went wrong while parsing XML content:", "E")
                        print(e)
                    else:
                        self.file_name = page
            else:
                _report_log(f"'{page}' does not exist.", "E")

    def _get_lines(self) -> list:
        """Get TextLine elements in PAGE XML Tree"""
        return [line for line in self.page_content.find_all("TextLine")]

    def _get_bounds_and_transcriptions(self) -> None:
        """Create bounds objects (for transcription) from baseline and boxes in XML tree"""
        self.bounds = []
        self.transcriptions = []
        self.pairs = []

        # TODO(@Alix) : add controls on these variables

        TEXT_DIRECTION = 'horizontal-lr'
        SCRIPT = 'default'

        for line in self._get_lines():
            try:
                baseline = _clean_coords(line.find("Baseline"))
            except AttributeError:
                # if the baseline is not good:
                baseline = None
            polygon = line.find("Coords")
            if polygon is not None:
                mask = _clean_coords(polygon)
            else:
                mask = []
            self.bounds.append({
                'lines': [{'baseline': baseline,
                           'boundary': mask,
                           'text_direction': TEXT_DIRECTION,
                           'script': SCRIPT}],  # self.document.main_script.name
                'type': 'baselines',
            })

            text_equiv = line.find('Unicode')
            if text_equiv:
                self.transcriptions.append(text_equiv.text)
            else:
                self.transcriptions.append('\n')

            self.pairs.append({'bounds': self.bounds[-1], 'transcription': self.transcriptions[-1]})

    def __init__(self, page):
        """Initialize a PageParser from a path to PAGE XML file or from the content of a PAGE XML file
        parsed with BeautifulSoup.

        :param page: either a path (str) either a parsed XML Tree
        :type page: str or BeautifulSoup
        """
        self.file_name = None
        self.page_content = None
        self._pcgts = None
        self.bounds = None
        self.transcriptions = None
        self.pairs = None
        self.source_image = None

        self._get_page_content(page)
        # proceed only if something was parsed
        if not self.page_content:
            _report_log("PageParser can't proceed: verify your input. Created an empty object.", "W")
        else:
            if self.page_content.find('PcGts'):
                self._pcgts = True
                creator = self.page_content.find('Creator')
                if not creator and not creator.text.lower() == 'escriptorium':
                    _report_log("It seems that the file was not created with" +
                                "eScriptorium, which may cause unexpected errors.", "W")
                image = self.page_content.find('Page')
                if image and 'imageFilename' in image.attrs:
                    self.source_image = image.attrs['imageFilename']
                else:
                    _report_log("There is no <Page> node with a @imageFilename" +
                                "attribute in this XML file. It can be an issue.", "W")
                try:
                    self._get_bounds_and_transcriptions()
                except Exception as e:
                    _report_log("Something went wrong while parsing TextLines", "E")
                    print(e)

            else:
                _report_log("Source is not a PAGE XML file", "E")
                _report_log("PageParser can't proceed: verify your input. Created an empty object.", "W")



# -*- coding: utf-8 -*-
# Authors : Alix Chagué <alix.chague@inria.fr>
#           Lucas Terriel <lucas.terriel@inria.fr>
# Licence : MIT
"""Client interface calls several sub-system of Kami-lib
"""
import os.path
from typing import Union

from kami.parser import (parser_text,
                         parser_page)
from kami.preprocessing.transformation import (RemoveDigits,
                                               ToLowerCase,
                                               ToUpperCase,
                                               RemovePunctuation,
                                               RemoveDiacritics,
                                               _Composer,
                                               count_diacritics)
from kami.transcription._base_prediction_io import (_load_model,
                                                    _load_image)
from kami.transcription.prediction import _Prediction
from kami.metrics.evaluation import Scorer


class Kami:
    """A Facade class provides a simple interface to the complex logic of one or
    several subsystems in Kami.

    This hide a complexity of Kami subsystems as :
    * Preprocessing
    * Prediction
    * Metrics
    """

    def __init__(self,
                 data: Union[str, list],
                 image="",
                 model="",
                 apply_transforms="",
                 verbosity=False,
                 truncate=False,
                 percent=False,
                 round_digits='.01'
                 ) -> None:
        # Inputs

        self.reference = None
        self.prediction = None
        self.model = None
        self.scores = None
        self.apply_transforms = [code for line in apply_transforms.split() for code in line]

        # Options

        self.truncate = truncate
        self.percent = percent
        self.verbosity = verbosity
        self.round_digits = round_digits

        # Preprocess sentences output
        self.reference_preprocess = ""
        self.prediction_preprocess = ""

        if isinstance(data, list) and len(data) > 1:
            # case with two text files => compute score
            if data[0].endswith('txt') and data[1].endswith('txt'):
                self.reference = parser_text._TextParser(data[0]).text
                self.prediction = parser_text._TextParser(data[1]).text
            # case with two strings => compute score
            else:
                self.reference = data[0]
                self.prediction = data[1]

            self.scores = Scorer(self.reference,
                                 self.prediction,
                                 truncate_score=self.truncate,
                                 show_percent=self.percent,
                                 round_digits=self.round_digits)

        # case with GT RAW TEXT file => create a HTR pipeline => compute scores
        elif isinstance(data, str) and data.endswith('txt'):
            self.reference = parser_text._TextParser(data).text
            self.file_name = os.path.basename(data)
            self.model = _load_model(model,
                                     verbosity=self.verbosity)
            self.image = _load_image(image,
                                     verbosity=self.verbosity)
            pipeline = _Prediction(self.file_name,
                                  self.image,
                                  self.model,
                                  code='txt',
                                  verbosity=self.verbosity)
            self.prediction = pipeline.get_transcription()
            self.scores = Scorer(self.reference,
                                 self.prediction,
                                 truncate_score=self.truncate,
                                 show_percent=self.percent,
                                 round_digits=self.round_digits)

        # case with GT XML PAGE => create a HTR pipeline => compute scores
        elif isinstance(data, str) and data.endswith('xml'):
            self.reference_parse = parser_page._PageParser(data)
            self.reference = "\n".join(self.reference_parse.transcriptions)
            self.file_name = os.path.basename(data)
            self.model = _load_model(model,
                                     verbosity=self.verbosity)
            self.image = _load_image(image,
                                     verbosity=self.verbosity)
            pipeline = _Prediction(self.file_name,
                                  self.image,
                                  self.model,
                                  reference=self.reference_parse,
                                  code='xml',
                                  verbosity=self.verbosity)
            self.prediction = pipeline.get_transcription()
            self.scores = Scorer(self.reference,
                                 self.prediction,
                                 truncate_score=self.truncate,
                                 show_percent=self.percent,
                                 round_digits=self.round_digits)


        else:
            raise ValueError("Something is wrong. Check your data (ground truth and/or prediction).")

        # Case with preprocessing and modulate .board dict of Scorer object
        if len(apply_transforms) > 0:
            # Create a new dict to save the different state of computations during the transformations
            new_score = dict()

            # Create a new list to save the functions of transformations
            to_compose = list()

            # A dictionary associate the user code / the preprocessing function / readable name of function
            # Codes legend for users :  D : digits / U : uppercase / L : lowercase / P : punctuation / X : diacritics
            CODES_TRANSFORMS = {
                "D": (RemoveDigits(), "non_digits"),
                "U": (ToUpperCase(), "uppercase"),
                "L": (ToLowerCase(), "lowercase"),
                "P": (RemovePunctuation(), "remove_punctuation"),
                "X": (RemoveDiacritics(), "remove_diacritics")
            }

            # Initialize the default dict that corresponding to computations on sequences before the applications of
            # transformations
            new_score["default"] = self.scores.board

            for code in CODES_TRANSFORMS.items():
                if code[0] in apply_transforms:
                    to_compose.append(code[1][0])
                    # Retrieve the scores of one transformation and add this in new dict with readable name
                    new_score[code[1][1]] = self._compute_state_transformations(code[1][0]).board
            # Retrieve the scores of all transformations in same time and add this in new dict with readable name
            scores_all_transforms, sequences_all_transforms = self._compute_all_transformations(to_compose)
            new_score["all_transforms"] = scores_all_transforms.board


            # Count total char transformed with
            # transformations in reference and prediction :
            # Add transformed sentences
            self.reference_preprocess = sequences_all_transforms[0]
            self.prediction_preprocess = sequences_all_transforms[1]


            if "D" or "P" in apply_transforms:
                # Compute the different lengths between string
                # (initial reference and transformed reference / initial reference and transformed prediction)
                # to compute the number of digits or punctuation remove
                total_char_removed_reference \
                    = (len(self.reference)
                        - len(self.reference_preprocess))
                total_char_removed_prediction \
                    = (len(self.prediction)
                        - len(self.prediction_preprocess))

                new_score["Total_char_removed_from_reference"] = total_char_removed_reference
                new_score["Total_char_removed_from_prediction"] = total_char_removed_prediction

            if "X" in apply_transforms:
                # Compute the number of diacritics removed
                total_diacritics_reference \
                    = (
                        count_diacritics(self.reference)
                        - count_diacritics(self.reference_preprocess))
                total_diacritics_prediction \
                    = (
                        count_diacritics(self.prediction)
                        - count_diacritics(self.prediction_preprocess))
                new_score["Total_diacritics_removed_from_reference"] = total_diacritics_reference
                new_score["Total_diacritics_removed_from_prediction"] = total_diacritics_prediction

            if "U" in apply_transforms:
                def count_lower(string):
                    total = sum(1 for c in string if c.islower())
                    return total
                # Compute total of lowercase char pass in uppercase in reference after transform
                total_lower_reference = count_lower(self.reference)
                total_lower_prediction = count_lower(self.prediction)
                new_score["Total_char_pass_in_uppercase_in_reference"] = total_lower_reference
                new_score["Total_char_pass_in_uppercase_in_prediction"] = total_lower_prediction

            if "L" in apply_transforms:
                def count_upper(string):
                    total = sum(1 for c in string if c.isupper())
                    return total
                # Compute total of lowercase char pass in uppercase in reference after transform
                total_upper_reference = count_upper(self.reference)
                total_upper_prediction = count_upper(self.prediction)
                new_score["Total_char_pass_in_lowercase_in_reference"] = total_upper_reference
                new_score["Total_char_pass_in_lowercase_in_prediction"] = total_upper_prediction

            if "D" or "P" or "L" or "U" or "X" in apply_transforms:
                new_score["Length_reference"] = len(self.reference)
                new_score["Length_prediction"] = len(self.prediction)
                new_score["Length_reference_transformed"] = len(self.reference_preprocess)
                new_score["Length_prediction_transformed"] = len(self.prediction_preprocess)

            # Add all scores to a final board
            self.scores.board = new_score

    def _compute_state_transformations(self, type_transform):
        """Compute scores for one transformation
        """
        transform = type_transform(
            [self.reference, self.prediction]
        )
        scores_transform = Scorer(
            transform[0],
            transform[1],
            truncate_score=self.truncate,
            show_percent=self.percent,
            round_digits=self.round_digits
        )

        return scores_transform

    def _compute_all_transformations(self, types_transforms):
        """Compute scores for all transformations
        """
        transform_for_all = _Composer(types_transforms)(
            [self.reference,
             self.prediction]
        )

        scores_transform_all = Scorer(
            transform_for_all[0],
            transform_for_all[1],
            truncate_score=self.truncate,
            show_percent=self.percent,
            round_digits=self.round_digits
        )

        return scores_transform_all, transform_for_all
